(**************************************************************************)
(*                           Structure de file                            *)
(**************************************************************************)

type 'a file == 'a list ;;

let (file_vide : 'a file) = [] ;;

let est_vide (f : 'a file) = f = [] ;;

let ajoute (f : 'a file) x = f @ [x] ;;

let retire (f : 'a file) =
  match f with
      [] -> failwith "file vide"
    | x::f' -> x,f'
;;

(* Il y a une facon bien plus efficace de representer les files, en utilisant
   deux pointeurs, l'un sur le premier element et l'un sur le dernier, les
   elements etant chaines par des pointeurs. Les operations sur la file
   se font alors en temps constant.
   Voir la bibliotheque Caml Light "queue.ml". *)

(**************************************************************************)
(*                    Parcours en largeur d'un graphe                     *)
(**************************************************************************)

type graphe == (int * int list) list ;;

let (g : graphe) = [ 1,[2;3] ; 2,[1;4;5] ; 3,[1] ; 4,[2] ;
		     5,[2] ; 6,[7] ; 7,[6] ];;

(* Version fonctionnelle, avec les files definies ci-dessus. *)

let ajoute_liste acc file liste =
  let rec aux file = function 
      [] -> file
    | s::reste -> if mem s acc then aux file reste
                               else aux (ajoute file s) reste
  in aux file liste
;;

let largeur (g : graphe) =
  let rec traite_sommet acc file =
      try 
	let s',file' = retire file in
	if mem s' acc then
	  traite_sommet acc file'
	else
	  let acc' = s'::acc in
	  let file'' = ajoute_liste (s'::acc) file' (assoc s' g) in
	  traite_sommet acc' file''
      with Failure "file vide" -> acc

  and traite_graphe acc = function
    []            -> acc
  | (s,al)::reste -> if mem s acc then
                       traite_graphe acc reste
                     else
                       let acc' = traite_sommet (s::acc) al in
                       traite_graphe acc' reste
  in
  rev (traite_graphe [] g)
;;

(* Version imperative, avec les files imperatives de Caml Light *)

let ajoute_liste acc file liste =
  let rec aux = function
      [] -> ()
    | s::reste -> if not (mem s acc) then queue__add s file
  in aux liste
;;

let largeur (g : graphe) =
  let file = queue__new () in
  let rec traite_sommet acc =
    try 
      let s = queue__take file in
      if not (mem s acc) then begin
	ajoute_liste (s::acc) file (assoc s g) ;
	traite_sommet (s::acc) 
      end else 
	traite_sommet acc
    with queue__Empty -> acc
  and traite_graphe acc = function
    []            -> acc
  | (s,al)::reste ->
      if mem s acc then
        traite_graphe acc reste
      else begin
	queue__clear file ;
	ajoute_liste (s::acc) file al ;
	let acc' = traite_sommet (s::acc) in
        traite_graphe acc' reste
      end 
  in
  rev (traite_graphe [] g)
;;
