(**************************************************************************)
(*                                Min Max                                 *)
(**************************************************************************)

(* minimize : ('a -> 'a -> bool) -> ('b -> 'a) -> 'b list -> 'b * 'a
 *
 * prend une fonction de comparaison, une fonction d'evaluation, une liste,
 * et rend un element de la liste minimizant la fonction d'evaluation,
 * ainsi que la valeur de la fonction d'evaluation en ce point. *)

let minimize cmp eval l =
  let rec trouve ((m,v) as c) = function
      []    -> c
    | x'::l -> let v' = eval x' in
	       if cmp v' v then
		 trouve (x',v') l
	       else 
		 trouve c l
  in match l with
      []   -> raise (Invalid_argument "minimize: empty list")
    | x::l -> trouve (x,eval x) l
;;

(* min et max dans le cas particulier d'evaluation dans float *)

let min (eval:'a->float) l = minimize (prefix <) eval l;;
let max (eval:'a->float) l = minimize (prefix >) eval l;;


(* minmax
     eval : 'a -> float
     suiv : 'a -> 'a list
     dep  : 'a
   renvoie le coup  jouer et son evaluation : 'a (pos. choisie) * float
   profondeur 1. *)

let minmax eval suiv dep =
  let eval1 p = match eval p with
      0.0 -> 1.0
    | 1.0 -> 0.0
    | _   -> snd (min eval (suiv p))
  in
    max eval1 (suiv dep)
;;


(* minmax_n : min-max profondeur n *)

let minmax_n eval suiv n dep =

  let rec eval1 n p = match eval p with
      0.0 -> 1.0
    | 1.0 -> 0.0
    | _   ->
	if n = 0 then
	  snd (min eval (suiv p))
	else
	  snd (min (fun p -> snd (minmax_rec (pred n) p)) (suiv p))

  and minmax_rec n p = 
    match eval p with
	(0.0 | 1.0) as e -> p,e
      | _                -> max (eval1 n) (suiv p)

  in minmax_rec n dep
;;

(* remarque : minmax = (minmax_n 0) *)
