(**************************************************************************)
(*                             Jeu du morpion                             *)
(**************************************************************************)

include "minmax";;

let taille = 4 ;;
let borne  = taille-1 ;;
let nb_victoire = 4 ;;

type joueur = JoueurA | JoueurB ;;

type case = Vide | JA | JB ;;

type configuration =
    { aqui   : joueur ;
      grille : case vect vect } ;;
      
let (configuration_initiale : configuration) =
  { aqui = JoueurA ;
    grille = make_matrix taille taille Vide }
;;

let affiche_configuration c =
  for j = 0 to borne do
    for i = 0 to borne do
      print_string (match c.grille.(i).(j) with
			Vide -> " ."
		      | JA   -> " A"
		      | JB   -> " B")
    done ;
  if j < borne then print_newline ()
  done ;
  print_string (match c.aqui with JoueurA -> "  Joueur A" 
		  | JoueurB -> "  Joueur B") ;
  print_newline ()
;;

let copy_grille g =
  let g' = make_matrix taille taille Vide in
    for i = 0 to borne do
      for j = 0 to borne do
	g'.(i).(j) <- g.(i).(j)
      done
    done ;
    g'
;;

let (suiv : configuration -> configuration list) = fun c ->
  let r = ref [] in
    for i = 0 to borne do
      for j = 0 to borne do
	match c.grille.(i).(j) with
	    Vide ->
	      let g = copy_grille c.grille in
	      r := (match c.aqui with
		     	JoueurA ->
			  g.(i).(j) <- JA ;
			  { aqui = JoueurB ; grille = g }
		      | JoueurB ->
			  g.(i).(j) <- JB ;
			  { aqui = JoueurA ; grille = g } ) :: !r
	  | _ -> ()
      done
    done ;
    !r
;;

let compte_align k jo grille =
  let cpt = ref 0 in
  for j = 0 to borne do
    for i = 0 to borne do
      if i <= taille-k then
      	if (try
	      for m = 0 to k-1 do
		if grille.(i+m).(j) <> jo then failwith "truc"
	      done ; true
	    with Failure "truc" -> false) then incr cpt ;
      if i <= taille-k  && j <= taille-k then
	if (try
	      for m = 0 to k-1 do
		if grille.(i+m).(j+m) <> jo then failwith "truc"
	      done ; true
	    with Failure "truc" -> false) then incr cpt ;
      if j <= taille-k then
	if (try
	      for m = 0 to k-1 do
		if grille.(i).(j+m) <> jo then failwith "truc"
	      done ; true
	    with Failure "truc" -> false) then incr cpt ;
       if i >=k-1  && j >= k-1 then
	if (try
	      for m = 0 to k-1 do
		if grille.(i-m).(j-m) <> jo then failwith "truc"
	      done ; true
	    with Failure "truc" -> false) then incr cpt
    done
  done ;
  !cpt
;;

let (eval : configuration -> float) = fun c ->
  let case_moi, case_autre = match c.aqui with
      JoueurA -> JA,JB
    | JoueurB -> JB,JA in
  if compte_align nb_victoire case_moi c.grille > 0 then
    1.0
  else if compte_align nb_victoire case_autre c.grille > 0 then
    0.0
  else
    let c_autre = compte_align (pred nb_victoire) case_autre c.grille in
      if c_autre > 0 then
	0.1
      else
    	let c_moi = compte_align (pred nb_victoire) case_moi c.grille in
	  0.1 +. 0.9 *. (float_of_int c_moi) /. 24.0
;;


let joue c = fst (minmax eval suiv c) ;;

let joue_n n c = fst (minmax_n eval suiv n c) ;;

let boucle coup ci =
  let rec boucle_rec c =
    if eval c = 0.0 then begin
      print_string (match c.aqui with JoueurA -> "le joueur B gagne !"
		                    | JoueurB -> "le joueur A gagne !") ;
      print_newline ()
    end else begin
      affiche_configuration c ;
      boucle_rec (coup c)
    end
  in
    boucle_rec ci
;;

