(**************************************************************************)
(*                                 LISTES                                 *)
(**************************************************************************)

(************************************)
(* fonctions simples sur les listes *)
(************************************)

let rec length = function
   []   -> 0
 | _::l -> 1 + length l;;

let rec somme = function
   []   -> 0
 | a::l -> a + somme l;;

let rec map f = function
   []   -> []
 | a::l -> (f a)::(map f l);;

let rec do_list f = function
   []   -> ()
 | a::l -> f a ; do_list f l;;

let rec concat l1 l2 = 
  match l1 with
     []   -> l2
   | a::l -> a::(concat l l2);;

let rec rev = function
   []   -> []
 | a::l -> concat (rev l) [a];;

let rev_lineaire l = 
  let rec aux acc = function
     []   -> acc 
   | a::l -> aux (a::acc) l
  in aux [] l;;

(*********************)
(* cas exceptionnels *)
(*********************)

let hd = function
   []   -> failwith "liste vide"
 | a::_ -> a;;

let tl = function
   []   -> failwith "liste vide"
 | _::l -> l;;

exception Not_found;;

let rec nth n = function
   []   -> raise Not_found
 | a::l -> if n=1 then a else nth (n-1) l;;


(**************************************)
(* iterateurs generaux sur les listes *)
(**************************************)

let rec iter_list f b = function
   []   -> b
 | a::l -> iter_list f (f b a) l;;

let rec list_iter f l b =
  match l with
     []    -> b
   | a::l' -> f a (list_iter f l' b);;


(***************************)
(* on peut alors redefinir *)
(***************************)

let length l  = iter_list (fun s _ -> succ s) 0 l;;
let somme   = iter_list (fun x y -> x+y) 0;;
let produit = iter_list (fun x y -> x*y) 1;;
let do_list f = iter_list (fun _ x -> (f x)) ();;
let rev l = iter_list (fun r x -> x::r) [] l;;

let map f l = list_iter (fun x r -> (f x)::r) l [];;
let concat l1 l2 = list_iter (fun x r -> x::r) l1 l2;;


(**************************************************************************)
(*                         QUICKSORT FONCTIONNEL                          *)
(**************************************************************************)

let pivot p l =
  let rec pivot_rec (acc1,acc2) = function
     []   -> acc1,acc2
   | a::l -> if a>p then pivot_rec (acc1,a::acc2) l
                    else pivot_rec (a::acc1,acc2) l
  in
  pivot_rec ([],[]) l
;;

let rec quicksort = function
    []   -> []
  | [x]  -> [x]
  | p::l -> let l1,l2 = pivot p l in (quicksort l1)@[p]@(quicksort l2)
;;


